// mouseinrect.cpp
// by Glenn G. Chappell
// October 2003
//
// For CS 381
// Demonstrates checking whether mouse coords lie inside
//  a screen-aligned rectangle.
// The interesting code is in the function "inrect".
// See also display, where glRectd is used for drawing.

#include <iostream>
using std::cerr;
using std::endl;
#include <stdlib.h>  // Some versions of MS-Vis C++ have broken <cstdlib>
//using std::exit;
#include <GL/glut.h> // GLUT stuff - includes OpenGL headers as well


// Global variables
// Window/viewport
const int startwinsize = 400; // Starting window width & height, in pixels

// Keyboard
const int ESCKEY = 27;        // ASCII value of escape character

// For rectangle
bool rect_is_red = false;     // true: rect colored red; false: rect colored blue
const double rectleft = -0.5;
const double rectright = 0.5;
const double rectbottom = -0.2;
const double recttop = 0.2;



// inrect
// Given a position (in the format that GLUT gives mouse positions)
//  and the boundaries of a screen-aligned rectangle, in 2-D object coordinates.
// Determine whether the position lies in the rectangle.
// Returns true if so, false if not.
// Assumes that the various transformations do not contain any rotations or shears;
//  only translation, scaling, and the usual projections are allowed.
bool inrect(int mx, int my,  // position, as GLUT gives mouse positions
            double left, double bottom, double right, double top) 
                             // bdaries of rectangle (object coord's)
{
   // GLUT likes y to increase top-to-bottom; OpenGL prefers bottom-to-top.
   // We convert GLUT-style y-coord "my" to OpenGL-style "my2".

   const int window_height = glutGet(GLUT_WINDOW_HEIGHT);  // Get window height
   const int my2 = window_height - my;                     // Convert "my" to zero-at-bottom

   // Now convert mx, my2 to eye coordinates in ex, ey.
   // Range is -1..+1 in both coords, with -1 being the left (resp. bottom) side
   //  of the viewport, and +1 being the right (resp. top).

   GLint viewport_dims[4];
   glGetIntegerv(GL_VIEWPORT, viewport_dims);  // Get viewport left, bottom, width, height
   const double ex = double(mx  - viewport_dims[0])/viewport_dims[2]*2.-1.;
   const double ey = double(my2 - viewport_dims[1])/viewport_dims[3]*2.-1.;

   // Get the model/view & projection transformation matrices.

   GLdouble modelviewd[16];
   glGetDoublev(GL_MODELVIEW_MATRIX, modelviewd);    // Get model/view matrix
   GLdouble projectiond[16];
   glGetDoublev(GL_PROJECTION_MATRIX, projectiond);  // Get projection matrix

   // Multiply the model/view & projection matrices together to obtain a single
   //  matrix representing the combination of the two transformations.
   // Store this matrix in the array "mymatrix".

   double mymatrix[4][4];  // Holds combination matrix
   int i, j, k;            // Loop counters
   for (i=0; i<4; ++i)
   {
      for (j=0; j<4; ++j)
      {
         double value = 0.;
         for (k=0; k<4; ++k)
         {
            value += modelviewd[i+4*k] * projectiond[k+4*j];
         }
         mymatrix[i][j] = value;
      }
   }

   // Convert the point (left, bottom) from object coordinates to eye coordinates:
   //  multiply the combination model/view+projection matrix by (left, bottom, 0).

   double vertin[4], vertout[4];  // Hold original & transformed vertices, respectively.
   vertin[0] = left;
   vertin[1] = bottom;
   vertin[2] = 0.;
   vertin[3] = 1.;
   for (i=0; i<4; ++i)
   {
      double value = 0.;
      for (j=0; j<4; ++j)
      {
         value += (mymatrix[i][j]*vertin[j]);
      }
      vertout[i] = value;
   }
   double eye_left   = vertout[0]/vertout[3];
   double eye_bottom = vertout[1]/vertout[3];

   // Convert the point (right, top) from object coordinates to eye coordinates:
   //  multiply the combination model/view+projection matrix by (right, top, 0).

   vertin[0] = right;
   vertin[1] = top;
   vertin[2] = 0.;
   vertin[3] = 1.;
   for (i=0; i<4; ++i)
   {
      double value = 0.;
      for (j=0; j<4; ++j)
      {
         value += (mymatrix[i][j]*vertin[j]);
      }
      vertout[i] = value;
   }
   const double eye_right  = vertout[0]/vertout[3];
   const double eye_top    = vertout[1]/vertout[3];

   // Now we have the given position and rectangle boundaries, all in eye coordinates.
   // Check whether the position lies inside the rectangle, and return the appropriate bool.
   // We check ">=" for one boundary, "<" for opposite boundary, to guarantee that
   //  the mouse will never lie inside both of two rectangles that share a boundary edge.

   return (ex >= eye_left
        && ex <  eye_right
        && ey >= eye_bottom
        && ey <  eye_top);
}


// display
// The GLUT display function
void display()
{
   glClear(GL_COLOR_BUFFER_BIT);

   if (rect_is_red)
      glColor3d(1., 0., 0.);
   else
      glColor3d(0., 0., 1.);
   glRectd(rectleft, rectbottom, rectright, recttop);

   glutSwapBuffers();
}


// motion
// The GLUT motion function
void motion(int x, int y)
{
   const bool new_rect_is_red = inrect(x, y,
                                       rectleft, rectbottom, rectright, recttop);
   if (new_rect_is_red != rect_is_red)
   {
      rect_is_red = new_rect_is_red;
      glutPostRedisplay();
   }
}


// passivemotion
// The GLUT passive motion function
void passivemotion(int x, int y)
{
   motion(x, y);  // Do the same thing as the motion function
}


// reshape
// The GLUT reshape function
void reshape(int w, int h)
{
    glViewport(0, 0, w, h);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // We set up coordinate system so that aspect ratios are always correct,
    //  and the region from -1..1 in x & y always just fits in the viewport.
    if (w > h)
    {
       gluOrtho2D(-double(w)/h, double(w)/h, -1., 1.);
    }
    else
    {
       gluOrtho2D( -1., 1., -double(h)/w, double(h)/w);
    }

    glMatrixMode(GL_MODELVIEW);  // Always go back to model/view mode
}


// keyboard
// The GLUT keyboard function
void keyboard(unsigned char key, int x, int y)
{
   switch (key)
   {
   case ESCKEY:  // ESC: Quit
      exit(0);
      break;
   }
}


// idle
// The GLUT idle function
void idle()
{
   // Print OpenGL errors, if there are any (for debugging)
   if (GLenum err = glGetError())
   {
      cerr << "OpenGL ERROR: " << gluErrorString(err) << endl;
   }
}


// init
// Initializes GL states
// Called by main
void init()
{
   glClearColor(1.0, 1.0, 1.0, 0.0);
}


int main(int argc, char ** argv)
{
   // Initialize OpenGL/GLUT
   glutInit(&argc, argv);
   glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);

   // Make a window
   glutInitWindowSize(startwinsize, startwinsize);
   glutInitWindowPosition(50, 50);
   glutCreateWindow("CS 381 - Checking mouse in rectangle (move the mouse)");

   // Initialize GL states & register callbacks
   init();
   glutDisplayFunc(display);
   glutMotionFunc(motion);
   glutPassiveMotionFunc(passivemotion);
   glutReshapeFunc(reshape);
   glutKeyboardFunc(keyboard);
   glutIdleFunc(idle);

   // Do something
   glutMainLoop();

   return 0;
}
